/* Program to control the system input hook.                    */

/* Copyright (C) 2002, David W Noon.  All rights reserved.      */

/* You may use this code freely. You may also distribute it     */
/* freely, provided the original copyright caveat is preserved  */
/* and no charge is levied beyond the price of its distribution */
/* medium.                                                      */

#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <process.h>
#ifdef __EMX__
#include <sys/types.h>
#include <dirent.h>
#else
#include <direct.h>
#endif

#define INCL_PM
#define INCL_WINSHELLDATA
#define INCL_WINWINDOWMGR
#define INCL_WINMESSAGEMGR
#define INCL_WINHOOKS
#define INCL_WINERRORS
#define INCL_DOSMODULEMGR
#define INCL_DOSPROCESS
#define INCL_DOSERRORS
#include <os2.h>

#include "ctlhook.h"

#ifndef __EMX__
#define _MAX_DIR   256  /* _MAX_DIR is a GNU special */
#define _MAX_FNAME 256  /* _MAX_FNAME is a GNU special */
#endif
static void make_ini_filename_from_exe(char ini_file[])
{
   PPIB pib_ptr;
   PTIB tib_ptr;
   char exe_name[CCHMAXPATH], exe_drive[_MAX_DRIVE], exe_path[_MAX_DIR], exe_basename[_MAX_FNAME];

   /* Obtain the .EXE name from OS/2 */
   DosGetInfoBlocks(&tib_ptr, &pib_ptr);
   DosQueryModuleName(pib_ptr->pib_hmte, sizeof exe_name, exe_name);

   /* Replace the .EXE extension with .INI, and that will do */
   _splitpath(exe_name, exe_drive, exe_path, exe_basename, NULL);
   _makepath(ini_file, exe_drive, exe_path, exe_basename, ".INI");

   return;
}

/* Communcations area for the threads controlling the input hook. */
typedef struct
{
   HWND    Dialogue_window;
   HWND    Object_window;
   HSWITCH Switch_entry;
   HMODULE Hook_dll;
   PFN     Hook_routine;
   BOOL    HookIsSet;
} HOOKCTL, * PHOOKCTL;

/* Message handler for object window. */
static MRESULT _System ObjectWndProc(HWND hWnd, ULONG msg, MPARAM mp1, MPARAM mp2)
{
   const PHOOKCTL hookctl_ptr = (PHOOKCTL) WinQueryWindowPtr(hWnd, 0L);
   const HAB anchor_block = WinQueryAnchorBlock(hWnd);
   APIRET rc;
   char crap[CCHMAXPATH];

   switch (msg)
   {
   case WM_USER_SETHOOK:
      if (!hookctl_ptr->HookIsSet)
      {
         rc = DosLoadModule(crap, sizeof crap, "INPTHOOK", &(hookctl_ptr->Hook_dll));
         if (rc == NO_ERROR)
         {
            /* Our hook is always at ordinal #1. */
            DosQueryProcAddr(hookctl_ptr->Hook_dll, 1UL, NULL, &(hookctl_ptr->Hook_routine));

            hookctl_ptr->HookIsSet = WinSetHook(anchor_block, NULLHANDLE, HK_INPUT,
                  hookctl_ptr->Hook_routine, hookctl_ptr->Hook_dll);
            if (hookctl_ptr->HookIsSet)
               WinPostMsg(hookctl_ptr->Dialogue_window, WM_USER_HOOKISSET,
                     MPFROMLONG(hookctl_ptr->Hook_dll),
                     MPFROMP(hookctl_ptr->Hook_routine));
         }
      }
      return MRFROMLONG(0L);
      break;

   case WM_USER_RELEASEHOOK:
      if (hookctl_ptr->HookIsSet)
      {
         if (WinReleaseHook(anchor_block, NULLHANDLE, HK_INPUT,
               hookctl_ptr->Hook_routine, hookctl_ptr->Hook_dll))
         {
            DosFreeModule(hookctl_ptr->Hook_dll);
            hookctl_ptr->HookIsSet = FALSE;
            hookctl_ptr->Hook_routine = NULL;
            hookctl_ptr->Hook_dll = NULLHANDLE;
            WinPostMsg(hookctl_ptr->Dialogue_window, WM_USER_HOOKISRELEASED, MPVOID, MPVOID);
         }
      }
      return MRFROMLONG(0L);
      break;

   case WM_USER_START:
      /* This is a reflection of WM_CREATE, but after the dispatch loop has begun. */
      if (hookctl_ptr->HookIsSet)
      {
         rc = DosLoadModule(crap, sizeof crap, "INPTHOOK", &(hookctl_ptr->Hook_dll));
         if (rc == NO_ERROR)
         {
            /* Our hook is always at ordinal #1. */
            DosQueryProcAddr(hookctl_ptr->Hook_dll, 1UL, NULL, &(hookctl_ptr->Hook_routine));

            hookctl_ptr->HookIsSet = WinSetHook(anchor_block, NULLHANDLE, HK_INPUT,
                  hookctl_ptr->Hook_routine, hookctl_ptr->Hook_dll);
         }
         else
            hookctl_ptr->HookIsSet = FALSE;
      }

      if (hookctl_ptr->HookIsSet)
         WinPostMsg(hookctl_ptr->Dialogue_window, WM_USER_HOOKISSET,
               MPFROMLONG(hookctl_ptr->Hook_dll),
               MPFROMP(hookctl_ptr->Hook_routine));
      else
         WinPostMsg(hookctl_ptr->Dialogue_window, WM_USER_HOOKISRELEASED, MPVOID, MPVOID);
      return MRFROMLONG(0L);
      break;

   case WM_CLOSE:
   case WM_DESTROY:
      if (hookctl_ptr != NULL && hookctl_ptr->HookIsSet)
      {
         WinReleaseHook(anchor_block, NULLHANDLE, HK_INPUT,
               hookctl_ptr->Hook_routine, hookctl_ptr->Hook_dll);
         DosFreeModule(hookctl_ptr->Hook_dll);
         hookctl_ptr->HookIsSet = FALSE;
         hookctl_ptr->Hook_routine = NULL;
         hookctl_ptr->Hook_dll = NULLHANDLE;
      }
      break;

   case WM_CREATE:
      /* Allow complete initialization to take place later. */
      WinPostMsg(hWnd, WM_USER_START, mp1, mp2);
      break;

   default:
      break;
   }

   /* Just the usual. */
   return WinDefWindowProc(hWnd, msg, mp1, mp2);
}

/* Main routine for the background thread that process object window messages. */
static void hookctl_thread(void * thread_parm)
{
   const PHOOKCTL hookctl_ptr = (PHOOKCTL) thread_parm;
   HAB   anchor_block;
   HMQ   msg_queue;
   HWND  Object_client_window;
   ULONG frame_ctl_flags = 0UL;
   QMSG  msg;
   static const UCHAR class_name[] = "object_window";

   anchor_block = WinInitialize(0UL);
   msg_queue = WinCreateMsgQueue(anchor_block, 0UL);

   /* Register a class name for the object window. */
   WinRegisterClass(anchor_block, class_name, &ObjectWndProc, 0, sizeof(PHOOKCTL));

   /* Create the object window -- basically just a message processor. */
   hookctl_ptr->Object_window = WinCreateStdWindow(HWND_OBJECT,
         0UL, &frame_ctl_flags, class_name, NULL,
         0UL, 0UL, 0UL, &Object_client_window);

   if (hookctl_ptr->Object_window != NULLHANDLE)
   {
      /* Let the object window's frame and client access the communications area. */
      WinSetWindowPtr(hookctl_ptr->Object_window, 0L, hookctl_ptr);
      WinSetWindowPtr(Object_client_window, 0L, hookctl_ptr);
   
      /* Now the message dispatch loop. */
      while (WinGetMsg(anchor_block, &msg, 0UL, 0UL, 0UL))
         WinDispatchMsg(anchor_block, &msg);
   }
   else
   {
      const ERRORID error_code = WinGetLastError(anchor_block);
      char msg_text[64];

      sprintf(msg_text, "Severity: %lu. Code: %lu", error_code >> 16, error_code & 0x0000FFFFUL);
      WinMessageBox(HWND_DESKTOP, NULLHANDLE, msg_text, "WinCreateStdWindow() error", 0UL,
            MB_APPLMODAL | MB_ERROR | MB_OK);
   }

   /* Clean up PM resources. */
   WinDestroyWindow(hookctl_ptr->Object_window);
   WinDestroyMsgQueue(msg_queue);
   WinTerminate(anchor_block);

   return;
}

/* Message processing routine for the visible dialogue. Doesn't do much, really. */
static MRESULT _System MainDlgProc(HWND hWnd, ULONG msg, MPARAM mp1, MPARAM mp2)
{
   const PHOOKCTL hookctl_ptr = (PHOOKCTL) WinQueryWindowPtr(hWnd, 0L);

   switch (msg)
   {
   case WM_USER_HOOKISSET:
      WinSetDlgItemText(hWnd, DID_STATUS, "Set");
      return MRFROMLONG(0L);
      break;

   case WM_USER_HOOKISRELEASED:
      WinSetDlgItemText(hWnd, DID_STATUS, "Released");
      return MRFROMLONG(0L);
      break;

   case WM_COMMAND:
      switch (SHORT1FROMMP(mp1))
      {
      case DID_TOGGLE:
         WinPostMsg(hookctl_ptr->Object_window,
               hookctl_ptr->HookIsSet ? WM_USER_RELEASEHOOK : WM_USER_SETHOOK,
               MPVOID, MPVOID);
         break;

      case DID_OK:
         WinPostMsg(hWnd, WM_CLOSE, MPVOID, MPVOID);
         break;

      default:
         break;
      }
      return MRFROMLONG(0L);
      break;

   case WM_CLOSE:
      {
         const HPOINTER app_icon = (HPOINTER) WinSendMsg(hWnd, WM_QUERYICON, MPVOID, MPVOID);

         if (app_icon != NULLHANDLE)
            WinDestroyPointer(app_icon);
      }

      /* Tell the object window to close, too. */
      WinPostMsg(hookctl_ptr->Object_window, WM_CLOSE, MPVOID, MPVOID);

      /* Take this window off public display. */
      if (hookctl_ptr->Switch_entry != NULLHANDLE)
      {
         WinRemoveSwitchEntry(hookctl_ptr->Switch_entry);
         hookctl_ptr->Switch_entry = NULLHANDLE;
      }
      break;

   case WM_INITDLG:
      {
         const PHOOKCTL hookctl_ptr = (PHOOKCTL) PVOIDFROMMP(mp2);
         PPIB pib_ptr;
         PTIB tib_ptr;
         HPOINTER app_icon;
         SWCNTRL switch_data;

         /* Save control area address for later. */
         WinSetWindowPtr(hWnd, 0L, hookctl_ptr);

         /* Load the program's icon as a resource. */
         app_icon = WinLoadPointer(HWND_DESKTOP, NULLHANDLE, 1UL);

         /* Set this window's icon to be the same. */
         if (app_icon != NULLHANDLE)
            WinPostMsg(hWnd, WM_SETICON, MPFROMLONG(app_icon), MPVOID);

         /* Add this baby to the PM shell's switch list. */
         DosGetInfoBlocks(&tib_ptr, &pib_ptr);
         memset(&switch_data, 0x00, sizeof switch_data);
         switch_data.hwnd = hWnd;
         switch_data.idProcess = pib_ptr->pib_ulpid;
         switch_data.uchVisibility = SWL_VISIBLE;
         switch_data.fbJump = SWL_JUMPABLE;
         switch_data.bProgType = PROG_PM;
         WinQueryWindowText(hWnd, sizeof(switch_data.szSwtitle), switch_data.szSwtitle);
         hookctl_ptr->Switch_entry = WinAddSwitchEntry(&switch_data);
      }
      return MRFROMLONG(0L);
      break;

   default:
      break;
   }

   /* Nothing special. */
   return WinDefDlgProc(hWnd, msg, mp1, mp2);
}

int main(int argc, char ** argv)
{
   HAB anchor_block;
   HMQ msg_queue;
   HOOKCTL comm_area;

   memset(&comm_area, 0x00, sizeof comm_area);

   /* See if the user has set a parameter. */
   if (argc > 1)
   {
      if (argc > 2 || (argv[1][0] != '-' && argv[1][0] != '/') || strlen(argv[1]) != 2)
         return 16;

      switch (argv[1][1])
      {
      case 's':
      case 'S':
         comm_area.HookIsSet = TRUE;
         break;

      case 'r':
      case 'R':
         comm_area.HookIsSet = FALSE;
         break;

      default:
         /* Bad parameter! */
         return 16;
         break;
      }
   }

   anchor_block = WinInitialize(0UL);
   msg_queue = WinCreateMsgQueue(anchor_block, 0L);

   /* Load the dialogue template */
   comm_area.Dialogue_window = WinLoadDlg(HWND_DESKTOP, NULLHANDLE,
         &MainDlgProc, NULLHANDLE, DID_MAIN_DIALOGUE, &comm_area);
   if (comm_area.Dialogue_window != NULLHANDLE)
   {
      HINI ini_handle;
      TID hookctl_thread_id;
      POINTL window_pos;
      ULONG buff_size = sizeof window_pos;
      SWP final_pos;
      char ini_filename[CCHMAXPATH];

      make_ini_filename_from_exe(ini_filename);
      ini_handle = PrfOpenProfile(anchor_block, ini_filename);
      if (ini_handle != NULLHANDLE)
      {
         /* Read any retained window position from last time */
         if (PrfQueryProfileData(ini_handle, "INPUT_HOOK", "WINDOW_ORIGIN",
               &window_pos, &buff_size) && buff_size == sizeof window_pos)
         {
            WinSetWindowPos(comm_area.Dialogue_window, HWND_TOP,
                  window_pos.x, window_pos.y, 0L, 0L, SWP_MOVE);
         }
         else
            window_pos.x = window_pos.y = LONG_MIN;
      }

      /* Start background thread. */
      hookctl_thread_id = _beginthread(&hookctl_thread, NULL, 48UL << 10, &comm_area);

      /* Display the dialogue */
      WinProcessDlg(comm_area.Dialogue_window);

      if (ini_handle != NULLHANDLE)
      {
         /* Obtain final window position for saving in the .INI file */
         WinQueryWindowPos(comm_area.Dialogue_window, &final_pos);

         /* Update the .INI only if the window position has changed */
         if (window_pos.x != final_pos.x || window_pos.y != final_pos.y)
         {
            window_pos.x = final_pos.x;
            window_pos.y = final_pos.y;
            PrfWriteProfileData(ini_handle, "INPUT_HOOK", "WINDOW_ORIGIN",
                  &window_pos, sizeof window_pos);
         }
         PrfCloseProfile(ini_handle);
      }

      /* The program has run its course. */
      WinDestroyWindow(comm_area.Dialogue_window);

      /* Wait for the worker thread to complete, if it hasn't already done so. */
      DosWaitThread(&hookctl_thread_id, DCWW_WAIT);
   }
   else
   {
      const ERRORID error_code = WinGetLastError(anchor_block);
      char msg_text[64];

      sprintf(msg_text, "Severity: %lu. Code: %lu", error_code >> 16, error_code & 0x0000FFFFUL);
      WinMessageBox(HWND_DESKTOP, NULLHANDLE, msg_text, "WinLoadDlg() error", 0UL,
            MB_APPLMODAL | MB_ERROR | MB_OK);
   }

   /* Finito, */
   WinDestroyMsgQueue(msg_queue);
   WinTerminate(anchor_block);

   return 0;
}
