/* The program generates all the scan codes from the keyboard.  */

/* Copyright (C) 2002, David W Noon.  All rights reserved.      */

/* You may use this code freely. You may also distribute it     */
/* freely, provided the original copyright caveat is preserved  */
/* and no charge is levied beyond the price of its distribution */
/* medium.                                                      */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#define INCL_NOPMAPI
#define INCL_DOSFILEMGR
#define INCL_DOSDEVICES
#define INCL_DOSDEVIOCTL
#define INCL_DOSNLS
#define INCL_DOSERRORS
#define INCL_KBD
#include <os2.h>

#include "scancode.h"

static const USHORT shift_state[7] = { 0U, 1U, 4U, 8U, 5U, 9U, 13U };

int main(void)
{
   SCANCODE scan_code[255]; /* scancodes go from 1 to keyboard maximum. */
   HFILE  kbd;
   APIRET rc;
   FILE * out_file;
   ULONG  code_page[3], code_page_size, open_action, param_size, data_size;
   USHORT sc, shift_patt;
   CPID   parameter_packet;
   struct
   {
      KBDKEYINFO key_info;
      USHORT     Kbd_DD_flags;
      USHORT     Xlate_flags;
      USHORT     Xlate_state1;
      USHORT     Xlate_state2;
   } data_packet;

   /* Start by getting the current code page. */
   rc = DosQueryCp(sizeof code_page, code_page, &code_page_size);
   if (rc == NO_ERROR)
   {
      memset(&parameter_packet, 0x00, sizeof parameter_packet);
      parameter_packet.idCodePage = code_page[0];

      /* Now open the keyboard for device access. */
      rc = DosOpen("KBD$", &kbd, &open_action, 0UL, FILE_READONLY,
            OPEN_ACTION_OPEN_IF_EXISTS,
            OPEN_SHARE_DENYNONE | OPEN_ACCESS_READONLY, NULL);
      if (rc == NO_ERROR)
      {
         /* Zero out everything. */
         memset(scan_code, 0x00, sizeof scan_code);

         for (sc = 0U; sc <= 254U; ++sc)
         {
            for (shift_patt = 0U; shift_patt < 7U; ++shift_patt)
            {
               memset(&data_packet, 0x00, sizeof data_packet);
               data_packet.key_info.chScan = sc + 1U;
               data_packet.key_info.fsState = shift_state[shift_patt];
   
               /* The docs say this IOCtl has to be done iteratively. */
               do
               {
                  rc = DosDevIOCtl(kbd, IOCTL_KEYBOARD, KBD_XLATESCAN,
                        &parameter_packet, sizeof parameter_packet, &param_size,
                        &data_packet, sizeof data_packet, &data_size);
               } while (rc == NO_ERROR && !(data_packet.Xlate_flags & 0x01));

               if (rc == NO_ERROR && !(data_packet.key_info.fbStatus & 0x01))
               {
                  scan_code[sc].scancode = sc + 1U;
                  if (data_packet.key_info.fbStatus & 0x02)
                  {
                     scan_code[sc].chr_vkey[shift_patt] = 2U; /* virtual key. */
                     scan_code[sc].chr_code[shift_patt] = data_packet.key_info.chScan;
                  }
                  else
                  {
                     scan_code[sc].chr_vkey[shift_patt] = 1U; /* character. */
                     scan_code[sc].chr_code[shift_patt] = data_packet.key_info.chChar;
                  }
               }
               else if (shift_patt == 0U) /* Bad scan code or just a shift key. */
                  break;
            }
         }

         /* Finished with keyboard. */
         DosClose(kbd);

         out_file = fopen("kbdscan.txt", "w");
         if (out_file != NULL)
         {
            shift_patt = 0U;

            for (sc = 0U; sc <= 254U; ++sc)
            {
               if (scan_code[sc].scancode != 0U)
               {
                  fprintf(out_file, "   { 0x%.02XU, { %u, %u, %u, %u, %u, %u, %u }, "
                     "{ 0x%.04XU, 0x%.04XU, 0x%.04XU, 0x%.04XU, 0x%.04XU, 0x%.04XU, 0x%.04XU } },\n",
                     scan_code[sc].scancode,
                     scan_code[sc].chr_vkey[0], scan_code[sc].chr_vkey[1],
                     scan_code[sc].chr_vkey[2], scan_code[sc].chr_vkey[3],
                     scan_code[sc].chr_vkey[4], scan_code[sc].chr_vkey[5],
                     scan_code[sc].chr_vkey[6],
                     scan_code[sc].chr_code[0], scan_code[sc].chr_code[1],
                     scan_code[sc].chr_code[2], scan_code[sc].chr_code[3],
                     scan_code[sc].chr_code[4], scan_code[sc].chr_code[5],
                     scan_code[sc].chr_code[6]);

                  ++shift_patt;
               }
            }
            fclose(out_file);

            printf("%u scan code entries were written to kbdscan.txt.", shift_patt);
         }
         else
            fputs("Unable to open file kbdscan.txt for output.", stderr);
      }
      else
         fprintf(stderr, "Unable to open keyboard. rc = %lu", rc);
   }
   else
      fprintf(stderr, "Error querying code page. rc = %lu", rc);

   return 0;
}
